#!/usr/bin/env perl

################################################################################
#
# Copyright (C) 2011 Sarah Eris Horsley Caffee
#
# This is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
#
# A limited rm wrapper.
#
# This scripts has 3 lists:
#     A blacklist of directories from which this script will absolutely refuse
#         to delete anything.  Individual files may be listed too.
#     A whitelist of directories from which deletions are always allowed.
#         Individual files may be listed too.
#     A whitelist of directories from which deletions are always allowed only if
#         they occur in subdirectories of the listed main directory.  Only
#         directories may be listed.  No individual files allowed here.
#
# Any file not specified as whitelisted or blacklisted will generate a
# warning prompt and offer the user a chance to cancel the deletion.
#
# The purpose is to help prevent accidental deletion of important system files.
#
# To use this, install this script somewhere in your path and add something 
# like the following to your default login scripts, such as the .bash_profile 
# file of the root user.
#
#    alias | grep -q "alias rm="
#    if [ $? -eq 0 ] ; then
#	RM_Opts=$(alias | awk '/alias rm=/ { sub(/^rm /, "", $2); print $2}' FS="'")
#    fi
#
#    unalias rm 2> /dev/null
#    alias rm="rm-limit.pl ${RM_Opts}"
#
# By installing this as an alias for rm that is set up in .bash_profile, it will
# only be active during interactive logins, and not when scripts are running.
#

# Last update: 2011/05/18 02:21:42

use strict;
use warnings;

use Cwd 'realpath';
use File::Basename;


################################################################################
#
# Note: / itself is protected by default. You are not allowed to delete the 
# entire filesystem using this script no matter what.
#
# Protecting or exposing a directory affects all subdirectories underneath it.
#
# The whitelist consists of directories from which we may always delete.

my @whitelist = (
    '/tmp',
    );

# whitelist_subdirs lists directories from which it is safe to delete only if
# we are deleting from a subdirectory of the listed directory.
# The purpose of this is to let us delete with impunity from user home directories
# but guard against accidental deletion of multiple home directories in their 
# entirety.

my @whitelist_subdirs = (
    '/home',
    );

# The blacklist consists of directories that we must never delete from
# under any circumstances.  To delete from these directories the user must
# invoke the /bin/rm command directly.

my @blacklist = (
    '/bin/',
    '/boot',
    '/etc',
    '/lib/',
    '/lib64',
    '/sbin',
    );

################################################################################

my $debug = 0;

my $rm="/bin/rm";
my $echo="/bin/echo";


my $proceed = "yes";
my $fail = 0;
my $file = 0;
my $opts = "";


setup_lists();


my $path = "";
foreach (@ARGV) {
    $path = $_;
    if ($path !~ /^-/) {
	if (check_whitelist($path) ) {
	    next;
	}

	if (check_whitelist_subdirs($path) ) {
	    next;
	}

	$proceed = "no";

	if (check_blacklist($path)) {
	    $fail = 1;
	}
    }
}

if ($fail) {
    print("File deletion aborted because blacklisted files were detected in the file list.
If you truly need to delete the files, please call /bin/rm directly.

");
    exit 1;
}



if ($proceed ne "yes") {
    print("

============ WARNING! ========== WARNING! ========== WARNING! =================

You are about to delete files or directories that are not in the whitelist of
safe locations from which to delete. Please review the rm command for any 
typos before proceeding.

Type \"yes\" to continue.


");
    $proceed = <STDIN>;
}

chomp($proceed);
if ($proceed eq "yes"){
    if ($debug) {
	exec($echo, ($rm, @ARGV));
    } else {
	exec($rm, @ARGV);
    }
} 

################################################################################
# Expand to full paths, append / to ends of directories.
# Returns empty string if the specified file does not exist.

sub normalize_name {
    my $path = $_;
    my $normal_path = "";

    if (-l $path) {
	$path = readlink($_);
    }

    $normal_path = realpath($path);
    if (!defined $normal_path) {
	return "";
    }

    if ((-d $normal_path) and ($normal_path !~ m{/$} )) {
	$normal_path = $normal_path."/";
    }

    return $normal_path;
}

################################################################################
# Return true if on whitelist

sub check_whitelist {
    my $regex = undef;
    my $path = normalize_name($_);
    for (my $i = 0; $i <= $#whitelist; $i += 1) {
	$regex = "^".quotemeta($whitelist[$i]);
	if ($whitelist[$i] !~ m{/$}) {
	    $regex = $regex."$$";
	}
	($debug) and print("regex is $regex\n");
	if (($path =~ $regex) and ($path ne $whitelist[$i])) {
	    $debug and print("Whitelisted for being in $whitelist[$i]: $_\n");
	    return 1;
	}
    }

    return 0;
}

################################################################################
# Return true if on whitelist_subdirs

sub check_whitelist_subdirs {
    my $regex = undef;
    my $path = normalize_name($_);
    for (my $i = 0; $i <= $#whitelist_subdirs; $i += 1) {
	$regex = "^".quotemeta($whitelist_subdirs[$i]);
	($debug) and print("regex is $regex\n");
	if (($path =~ $regex) and ($path ne $whitelist_subdirs[$i])) {
	    $regex = $regex.".*/.+";
	    ($debug) and print("new regex is $regex\n");
	    if ($path =~ $regex) {
		$debug and print("Whitelisted for being subdir of $whitelist_subdirs[$i]: $_\n");
		return 1;
	    }
	}
    }

    return 0;
}

################################################################################
# Return true if on blacklist

sub check_blacklist {
    my $regex = undef;
    my $path = normalize_name($_);

    # Always blacklist /
    if ($path eq "/\n") {
	print("Blacklisted for being the entire system: /");
	return 1;
    }

    for (my $i = 0; $i <= $#blacklist; $i += 1) {
	$regex = "^".quotemeta($blacklist[$i]);
	if ($blacklist[$i] !~ m{/$}) {
	    $regex = $regex."$$";
	}
	($debug) and print("regex is $regex\n");
	if ($path =~ $regex) {
	    print("Blacklisted for being in $blacklist[$i]: $_\n");
	    return 1;
	}
    }

    return 0;
}

################################################################################
# At the moment all this does is make sure that directories listed in the 
# lists all have a / at the end.

sub setup_lists {
    for (my $i = 0; $i <= $#whitelist; $i += 1) {
	if (-d $whitelist[$i]  and $whitelist[$i] !~ m{/$}) {
	    $whitelist[$i] = $whitelist[$i]."/";
	}
    }

    # All entries on the whitelist_subdirs list _must_ be directories.
    for (my $i = 0; $i <= $#whitelist_subdirs; $i += 1) {
	if ($whitelist_subdirs[$i] !~ m{/$}) {
	    $whitelist_subdirs[$i] = $whitelist_subdirs[$i]."/";
	}
    }
    for (my $i = 0; $i <= $#blacklist; $i += 1) {
	if (-d $blacklist[$i]  and $blacklist[$i] !~ m{/$}) {
	    $blacklist[$i] = $blacklist[$i]."/";
	}
    }

    if ($debug) {
	print("whitelist:\n");
	for (my $i = 0; $i <= $#whitelist; $i += 1) {
	    print($whitelist[$i]."\n");
	}
	print("whitelist_subdirs:\n");
	for (my $i = 0; $i <= $#whitelist_subdirs; $i += 1) {
	    print($whitelist_subdirs[$i]."\n");
	}
	print("blacklist:\n");
	for (my $i = 0; $i <= $#blacklist; $i += 1) {
	    print($blacklist[$i]."\n");
	}
    }
  }
